#include "tangentSpace.h"

namespace SceneManagement
{

void TangentSpace::createTangentSpaceVectors(Vector3f *v1, Vector3f *v2, Vector3f *v3,  //Vektory vrcholu
										float v1s, float v1t,					//texturove coordinaty 1. vrcholu
										float v2s, float v2t,					//texturove koordinaty 2. vrcholu
										float v3s, float v3t,					//texturove koordinaty 3. vrcholu
										Vector3f *vTangent,
										Vector3f *vBiNormal,
										Vector3f *vNormal)
{
	// Create edge vectors from vertex 1 to vectors 2 and 3.
    Vector3f edge_v2_v1 = *v2 - *v1;
    Vector3f edge_v3_v1 = *v3 - *v1;

    // Create edge vectors from the texture coordinates of vertex 1 to vector 2 and 3
	Vector2f tex_v2_v1(v2s-v1s, v2t-v1t);
	Vector2f tex_v3_v1(v3s-v1s, v3t-v1t);

    float fDenominator = tex_v2_v1.s * tex_v3_v1.t - tex_v3_v1.s * tex_v2_v1.t;

    if( fDenominator < 0.0001f && fDenominator > -0.0001f )
    {
        // We're too close to zero and we're at risk of a divide-by-zero! 
        // Set the tangent matrix to the identity matrix and do nothing.
        *vTangent  = Vector3f( 1.0f, 0.0f, 0.0f );
        *vBiNormal = Vector3f( 0.0f, 1.0f, 0.0f );
        *vNormal   = Vector3f( 0.0f, 0.0f, 1.0f );
    }
    else
    {
        // Calculate and cache the reciprocal value
        float fScale1 = 1.0f / fDenominator;

        Vector3f T;
        Vector3f B;
        Vector3f N;

        T = Vector3f((tex_v3_v1.t * edge_v2_v1.x - tex_v2_v1.t * edge_v3_v1.x) * fScale1,
                     (tex_v3_v1.t * edge_v2_v1.y - tex_v2_v1.t * edge_v3_v1.y) * fScale1,
                     (tex_v3_v1.t * edge_v2_v1.z - tex_v2_v1.t * edge_v3_v1.z) * fScale1);

        B = Vector3f((-tex_v3_v1.s * edge_v2_v1.x + tex_v2_v1.s * edge_v3_v1.x) * fScale1,
                     (-tex_v3_v1.s * edge_v2_v1.y + tex_v2_v1.s * edge_v3_v1.y) * fScale1,
                     (-tex_v3_v1.s * edge_v2_v1.z + tex_v2_v1.s * edge_v3_v1.z) * fScale1);

        // The normal N is calculated as the cross product between T and B
		N = T.crossProduct(B);

        // Calculate and cache the reciprocal value
        float fScale2 = 1.0f / ((T.x * B.y * N.z - T.z * B.y * N.x) + 
                                (B.x * N.y * T.z - B.z * N.y * T.x) + 
                                (N.x * T.y * B.z - N.z * T.y * B.x));

        //
        // Use the temporary T (Tangent), (B) Binormal, and N (Normal) vectors 
        // to calculate the inverse of the tangent matrix that they represent.
        // The inverse of the tangent matrix is what we want since we need that
        // to transform the light's vector into tangent-space.
        //

        (*vTangent).x =  T.x * fScale2;
        (*vTangent).y = -(B.x * fScale2);
        (*vTangent).z =  N.x * fScale2;
        (*vTangent).normalize();

        (*vBiNormal).x = -(T.y * fScale2);
        (*vBiNormal).y =  B.y * fScale2;
        (*vBiNormal).z = -(N.y * fScale2);
        (*vBiNormal).normalize();

        (*vNormal).x =  T.z * fScale2;
        (*vNormal).y = -(B.z * fScale2);
        (*vNormal).z =  N.z * fScale2;
        (*vNormal).normalize();

        //
        // NOTE: Since the texture-space of Direct3D and OpenGL are laid-out 
        //       differently, a single normal map can't look right in both 
        //       unless you make some adjustments somewhere.
        //
        //       You can adjust or fix this problem in three ways:
        //
        //       1. Create two normal maps: one for OpenGL and one for Direct3D.
        //       2. Flip the normal map image over as you load it into a texture 
        //          object.
        //       3. Flip the binormal over when computing the tangent-space
        //          matrix.
        //
        // Since the normal map used by this sample was created for Direct3D,
        // I've decided to simply flip the binormal.
        //
        *vBiNormal = *vBiNormal * -1.0f;
    }
}
}